//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//% TEXT2CLP.C - Copy the contents of a named file to the Windows Clipboard %
//% Written by James Greene, Schliengen, Germany                            %
//%-------------------------------------------------------------------------%
//% Last Modified:                                                          %
//%     20.01.1999 - Creation                                               %
//%     5.02.2003 - Fixed an off-by-one bug reported by						%
//%					Mikael Karlsson [mikael.karlsson@pitch.se]				%
//%     06.01.2015 - Recompiled with mingw / gcc 4.8.1                      %
//%		12.11.2016 - Added support for additional formats (HTML, RTF, etc.) %
//%-------------------------------------------------------------------------%
//% TEXT2CLP Version 2.0 (November 2016)                                    %
//% Copyright (c) 1999, 2003, 2015, 2016 by James Greene					%
//% TEXT2CLP comes with NO WARRANTY to the extent permitted by law.         %
//% You may redistribute copies of TEXT2CLP under the terms of the          %
//% GNU General Public License. For more information about these            %
//% matters, see the file named COPYING, or contact the author at           %
//% <james.greene@outlook.com>                                              %
//%                                                                         %
//% Please report any bugs to <james.greene@outlook.com>                    %
//%                                                                         %
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <conio.h>
#include <errno.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <fcntl.h>
#include <ctype.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#define VER_NUMBER  "2.01"
#define VER_STRING  "Version " VER_NUMBER
#define VER_DATE    "06. December 2016"

static char *strVer[] = {
//	"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
	"Copyright \xb8 1999, 2003, 2015, 2016 James Greene. All rights reserved.",
	"CLIP2TXT comes with NO WARRANTY to the extent permitted by law. You may",
	"redistribute copies of CLIP2TXT under the terms of the GNU General Public",
	"License. For more information about these matters, see the file named",
	"COPYING.txt, or contact the author at <james.greene@outlook.com>.",
	"",
	"Please report any bugs to <james.greene@outlook.com>.",
	""
};

#define	numVer (sizeof(strVer) / sizeof(char *))

static char *strUsage[] = {
		//	"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
	    "Options:  ( /option or --option are also accepted )",
	    "  -D[isplay]   = Display the Clipboard contents on the screen.",
		"  -F[ormat]=[xyz] where [xyz] = ASC or DOS, ANS or TXT, HTM, UNI, RTF or C2T.",
		"                  for ASCII, ANSI, HTML, UNICODE, Rich Text Format or CLIP2Txt.",
		"  -A[ppend]    = Append the file to the Clipboard (rather than replacing).",
	    "  -H[elp]      = Display this usage information.",
	    "  -?           = Same as -help.",
	    "  -N[ologo]    = Don't display the Copyright banner.",
	    "  -Q[uiet]     = Same as -nologo.",
	    "  -S[ilent]    = Same as -nologo.",
	    "  -T[ranslate] = (depreciated) Same as -F:DOS.",
		"",
	    "  The following options are all interpreted the same:",
	    "     -d, /d, -DISP, /Disp, --DISPLAY, -dIsP, ... you get the idea!\n"
};

#define	numUsage (sizeof(strUsage) / sizeof(char *))

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Clipboard Format options (used by the command-line processor)

#define CF_C2TFORM	0xF045		// proprietary CLIP2Txt Clipboard dump file

#define idxHTM 	0
#define idxRTF 	1
#define idxTXT  2
#define idxDOS  4
#define idxUNI	6
#define idxC2T	7

static struct cfOPT {
	unsigned cfVal;
	unsigned cfInt;
	char *cfName;
	char *cfDisp;
} cfValues[] = {
	{0, idxHTM, "HTM", "HTML Format"},			// HTM = 0
	{0, idxRTF, "RTF", "Rich Text Format"},			// RTF = 1
	{CF_TEXT, idxTXT, "ANS", "Text Format"},
	{CF_TEXT, idxTXT, "TXT", "Text Format"},
	{CF_OEMTEXT, idxDOS, "ASC", "OEM Text Format"},
	{CF_OEMTEXT, idxDOS, "DOS", "OEM Text Format"},
	{CF_UNICODETEXT, idxUNI, "UNI", "Unicode Text Format"},
	{CF_C2TFORM, idxC2T, "C2T", "CLIP2Txt Format"},
};

#define numCFOPT	(sizeof(cfValues) / sizeof(struct cfOPT))


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

typedef struct tagOPTIONS {
    unsigned fQuiet: 1;
    unsigned fDisplay: 1;
    unsigned fHelp: 1;
    unsigned fAppend: 1;
    unsigned fVersion: 1;
    unsigned uFormat;
    unsigned cfHTML;
    unsigned cfRTF;
    char szName[MAX_PATH];
} OPTIONS;

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  void    AnyKey(void) {
    register int ri;
    char *szMsg;

    szMsg = "(press any key to continue)";
    
    fprintf(stderr, szMsg);
    fflush(stderr);
    
    if (!getch())
        getch();
        
    for (ri=strlen(szMsg) ; ri > 0; ri--) 
        fprintf(stderr, "\b \b");
    fflush(stderr);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static void Version(void) {
	int ri;
    fprintf(stderr, "TEXT2CLP Version %s (%s at %.5s)\n", VER_NUMBER, VER_DATE, __TIME__);
    for (ri = 0; ri < numVer; ri++) {
    	fputs(strVer[ri], stderr);
    	fputs("\n", stderr);
    }
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  void    Usage(char *exe) {
	int ri;
    fprintf(stderr, "Usage: %s [options] filename [options]\n\n", exe);
    
    for (ri = 0; ri < numUsage; ri++) {
    	fputs(strUsage[ri], stderr);
    	fputs("\n", stderr);
    }

    AnyKey();
    
    //				"....-....1....-....2....-....3....-....4....-....5....-....6....-....7....-....8",
    fprintf(stderr, "Examples:\n");
    fprintf(stderr, "  %s FOO.TXT\n", exe);
    fprintf(stderr, "    Copies the contents of FOO.TXT to the Windows Clipboard.\n");
    fprintf(stderr, "  %s -f:htm \"C:\\My Documents\\Long File Name.html\"\n", exe);
    fprintf(stderr, "    Copies the contents of the specified file to the Clipboard as HTML-\n    formatted text.\n");
    fprintf(stderr, "    Filenames containing spaces must be \"enclosed in double quotes.\"\n\n");

    Version();
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  BOOL ParseCommandLine(char *argv[], OPTIONS *opt) {
    char *pgm, *arg;
    register int ri;
    char *MoreInfo;

    MoreInfo = "Try `TEXT2CLP --help' for more information.\n";

    memset(opt, 0, sizeof(OPTIONS));
    opt->uFormat = CF_TEXT;

    // query Clipboard formats
	opt->cfRTF = RegisterClipboardFormat("Rich Text Format");
	opt->cfHTML = RegisterClipboardFormat("HTML Format");

	// update the cfValues array
	cfValues[idxHTM].cfVal = opt->cfHTML;
	cfValues[idxRTF].cfVal = opt->cfRTF;

    pgm = *argv;
    
    while ((arg = *++argv) != NULL) {
        if ((*arg == '/') || (*arg == '-')) {   // flag
            ri = strlen(++arg);
            if (*arg == '-')
                ri = strlen(++arg);
          
            if (toupper(*arg) == 'F') {
            	char *brk = strpbrk(arg, ":=");
            	if (brk != 0) {
            		int fri = brk - arg;
            		++brk;

            		if (_strnicmp(arg, "FORMAT", fri) == 0) {
            			int lv;
            			int savFormat = opt->uFormat;
            			opt->uFormat = 999999;
            			for (lv = 0; (lv < numCFOPT) ; lv++) {
            				if (_strnicmp(brk, cfValues[lv].cfName, 3) == 0) {
            					opt->uFormat = cfValues[lv].cfVal;
            					break;
            				}
            			}

            			if (opt->uFormat == 999999) {
            				fprintf(stderr, "TEXT2CLP: Unrecognized format '%s'\n", brk);
            				opt->uFormat = savFormat;
            			}
            		}
            	} else {
                    fprintf(stderr, "TEXT2CLP: Option '%s' not recognized - ignored\n", --arg);
            	}
            }
            else if (_strnicmp(arg, "TRANSLATE", ri) == 0) {
            	if (opt->uFormat == CF_TEXT)
            		opt->uFormat = CF_OEMTEXT;
            }
            else if (_strnicmp(arg, "DISPLAY", ri) == 0)
                opt->fDisplay = TRUE;
            else if (_strnicmp(arg, "APPEND", ri) == 0)
                opt->fAppend = TRUE;
            else if ((_strnicmp(arg, "NOLOGO", ri) == 0) ||
                      (_strnicmp(arg, "QUIET", ri) == 0) ||
                      (_strnicmp(arg, "SILENT", ri) == 0))
                opt->fQuiet = TRUE;
            else if ((_strnicmp(arg, "HELP", ri) == 0) ||
                      (_strnicmp(arg, "?", ri) == 0))
                opt->fHelp = TRUE;
            else if ((_strnicmp(arg, "VERSION", ri) == 0))
                opt->fVersion = TRUE;
            else
                fprintf(stderr, "TEXT2CLP: Option '%s' not recognized\n%s", --arg, MoreInfo);
        } else {
            if (!*(opt->szName))
                strcpy(opt->szName, arg);
            else
                fprintf(stderr, "TEXT2CLP: Too many parameters: '%s' ignored\n%s", arg, MoreInfo);
        }
    }

    if (!opt->fQuiet) {
        fprintf(stderr, "%-64s %14s\n", "TEXT2CLP: Copy a text file to the Windows Clipboard", VER_STRING);
        fprintf(stderr, "Copyright \xb8 1999, 2003, 2015, 2016 by James Greene. All rights reserved.\nASP Member #06128.\n");
        fprintf(stderr, "%38s * %-38s\n\n", "http://informatics-consulting.de", "james.greene@outlook.com");
    }

    if (opt->fHelp) {
        Usage(pgm);
        return FALSE;
    }

    if (opt->fVersion) 
        Version();
        
    // if no filename specified, then just display the text to the screen
    if (!*(opt->szName)) {
        fprintf(stderr, "TEXT2CLP: No filename specified.\n%s", MoreInfo);
        return FALSE;
    }

    return TRUE;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// assumptions:

static	char	*HTMLHeader(OPTIONS *opt, DWORD dwSize) {
	char *strFmt = "Version:0.9\r\nStartHTML:%010u\r\nEndHTML:%010lu\r\nStartFragment:%010u\r\nEndFragment:%010lu\r\nSourceURL:file:///%s\r\n";
	unsigned uWid = strlen(strFmt) + 16 + strlen(opt->szName);		// 18 (2x4 and 2x5) because %010u/%010lu takes 5 or 6 chars, while the result 0000000001 needs 5 or 4 additional chars. The %s (2 bytes) is the buffer for the null at the end of the string.
	char *pHdr = malloc(uWid);

	if (pHdr)
		sprintf(pHdr, strFmt, uWid+1, dwSize+uWid+1, uWid+1, dwSize+uWid+1, opt->szName);

	return pHdr;
}


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

char *FmtToName(UINT uFmt) {
	int ri;

	for (ri = 0; ri < numCFOPT; ri++) {
		if (cfValues[ri].cfInt == uFmt)
			return cfValues[ri].cfDisp;
	}

	return "(unknown)";
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// This function is called if the "C2T" dump format was specified. In this case,
// we'll read a binary file which was created with CLIP2Txt and try to restore
// each of the formats it contains.

#define C2TSIG	0xA6543243		// "C2T"
struct _c2t_seg	{
	DWORD dwSig;
	unsigned uFmt;
	DWORD dwStart;
	DWORD dwLen;
};

static BOOL		CopyC2T(OPTIONS *opt) {
	struct _c2t_seg c2tseg;
	int fh;
	int ri;
	long lpos;
	HGLOBAL nd;
	LPSTR ns;
	int rdct;

	if (opt->fAppend)
        fprintf(stderr, "TEXT2CLP: -APPEND is not compatible with -FORMAT:C2T\n");

	// the C2T file format: c2t_seg header | binary data | c2t_seg header | binary data ...
	fh = _open(opt->szName, O_RDONLY | O_BINARY);
    if (fh == -1) {
        fprintf(stderr, "TEXT2CLP: Unable to open file %s for reading: %s\n",
                opt->szName, strerror(errno));
        return FALSE;
    }

    ri = _read(fh, &c2tseg, sizeof(c2tseg));
    if ((ri != sizeof(c2tseg)) || (c2tseg.dwSig != C2TSIG)) {	// no tag, probably not a C2T file
    	fprintf(stderr, "TEXT2CLP: %s does not appear to be a valid C2T (CLIP2TXT dump) file.\n", opt->szName);
    	_close(fh);
    	return FALSE;
    } else
    	lpos = _lseek(fh, 0, SEEK_SET);

    if (OpenClipboard(NULL)) {
    	EmptyClipboard();
        CloseClipboard();
    } else {
    	fprintf(stderr, "TEXT2CLP: Error accessing the Windows Clipboard.\n");
    	_close(fh);
    	return FALSE;
    }

    lpos = _tell(fh);
    while ((ri = _read(fh, &c2tseg, sizeof(c2tseg))) > 0) {
    	if (c2tseg.dwSig != C2TSIG) {	// no tag, so step forward 1 byte and try again.
    		lpos = _lseek(fh, lpos+1, SEEK_SET);
#ifdef _DEBUG
printf("c2tseg fault: offset %lx - want %x - got %lx\n", lpos, C2TSIG, c2tseg.dwSig);
#endif
    		break;
    	}

#ifdef _DEBUG
printf("c2tseg.%s = %lx\n", "dwSig", c2tseg.dwSig);
printf("c2tseg.%s = %lx\n", "dwStart", c2tseg.dwStart);
printf("c2tseg.%s = %lx\n", "dwLen", c2tseg.dwLen);
printf("c2tseg.%s = %x\n", "uFmt", c2tseg.uFmt);
#endif

    // we've got a valid header, so allocate enough space and read in the data

        nd = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, c2tseg.dwLen+2);
        if (nd != NULL) {
        	ns = (LPSTR)GlobalLock(nd);
            if (ns != NULL) {
            	rdct = _read(fh, ns, (UINT)c2tseg.dwLen);
            	ns[rdct+1] = '\0';
            	GlobalUnlock(nd);

            	if (OpenClipboard(NULL)) {
                    SetClipboardData(cfValues[c2tseg.uFmt].cfVal, nd);
                    CloseClipboard();

                    if (!opt->fQuiet)
                    	fprintf(stderr, "TEXT2CLP: %u bytes (%s) written to the Clipboard\n", (unsigned)rdct, FmtToName(c2tseg.uFmt));
                } else {
                	GlobalFree(nd);
                    fprintf(stderr, "TEXT2CLP: Unable to open the Windows Clipboard!\n");
                }
            } else
            	GlobalFree(nd);
    	} else {
            fprintf(stderr, "TEXT2CLP: Can't allocate %ld bytes memory\n", c2tseg.dwLen+2);
    	}


        lpos = _tell(fh);
    }

    _close(fh);
    return TRUE;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
static  BOOL    ClipCopy(OPTIONS *opt) {
    HANDLE nd;
    LPSTR ns;
    register int ri;
    int fh;
    struct _stat buf;
    DWORD dwSize, dwSeg;
    int nRest;
#define SEGMENT     16384
    BOOL rv = FALSE;
    char *htmlhead = "";
    unsigned uhead = 0;

    if (!(*opt->szName)) {                          // this should never happen
        fprintf(stderr, "TEXT2CLP: No filename specified.\n");
        return FALSE;
    }
        
    if (_stat(opt->szName, &buf) != -1) {
        dwSize = buf.st_size;
    } else {
        fprintf(stderr, "TEXT2CLP: Can't read %s: %s\n", opt->szName, strerror(errno));
        return FALSE;
    }

    if (opt->uFormat == CF_C2TFORM)
    	return CopyC2T(opt);

    if (opt->uFormat == opt->cfHTML) {
    	htmlhead = HTMLHeader(opt, dwSize);
    	uhead = strlen(htmlhead);
    }

// allocate some memory for the file    

	// 05. Feb. 2003 - Added +2 to the memory allocation - 
	//		without it, there was a risk of the loss of the final character from the file.

    nd = GlobalAlloc(GMEM_MOVEABLE|GMEM_DDESHARE, dwSize+2+uhead);
    if (nd != NULL) {
    	ns = (LPSTR)GlobalLock(nd);
        if (ns != NULL) {

// open the specified file

        	fh = _open(opt->szName, O_RDONLY | O_BINARY);
            if (fh == -1) {
                fprintf(stderr, "TEXT2CLP: Unable to open file %s for reading: %s\n",
                        opt->szName, strerror(errno));
                *ns = '\0';
            } else {
            	if (uhead) {
					memcpy(ns, htmlhead, uhead);
					ns += uhead;
					free(htmlhead);
					htmlhead = "";
            	}

                dwSeg = 0L;
                nRest = min(SEGMENT, dwSize);
                
                while (nRest) {
                    if ((ri = read(fh, ns, nRest)) < 0) {
                        fprintf(stderr, "TEXT2CLP: Unable to read file %s: %s\n",
                            opt->szName, strerror(errno));
                        break;
                    }

                    ns[ri] = '\0';
                            
                    if (opt->fDisplay)
                        fputs(ns, stdout);
                        
                    ns += ri;
                    dwSeg += ri;
                    nRest = min(nRest, (int)(dwSize - dwSeg));
                }
                
                close(fh);
            }

            GlobalUnlock(nd);

            if (OpenClipboard(NULL)) {
            	if (!opt->fAppend)
            		EmptyClipboard();
                SetClipboardData(opt->uFormat, nd);
                CloseClipboard();
    
                if (!opt->fQuiet) 
                    fprintf(stderr, "TEXT2CLP: %lu bytes copied to the Windows Clipboard.\n", dwSize);

                rv = TRUE;
            } else 
                fprintf(stderr, "TEXT2CLP: Unable to open the Windows Clipboard!\n");
        }
    }
    else
        fprintf(stderr, "TEXT2CLP: Can't allocate %ld bytes memory\n", dwSize);

    return rv;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
int main(int argc, char *argv[]) {
    OPTIONS opt;

    return ParseCommandLine(argv, &opt) && ClipCopy(&opt) ? 1 : 0;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

